/*
 * Copyright: Nicholas P Wiebe (2008-2009) and Irmtraud M Meyer (2008-2009)
 * License: licensed under the GNU General Public License version 3 (GPLv3)
 * AlignedHelix.h
 * Class for storing a helix as a set of base-pairs
 */

#ifndef ALIGNEDHELIX_H_
#define ALIGNEDHELIX_H_

#include <vector>
#include <set>
#include <string>
#include "Helix.h"
#include "Tree.h"
#include "BigDouble.h"

using namespace std;

class Alignment;

enum HelixPosition {LEFT_OF_BOUNDARY, RIGHT_OF_BOUNDARY, OVERLAPPING_BOUNDARY, INVALID_HELIX};

class AlignedHelix {
public:
	AlignedHelix(Alignment* align, const Helix & helix, int seqIndex, vector<int> & seq2AlignmentMap);
	AlignedHelix(Alignment* align, int outerBpPos5, int outerBpPos3, int length); // for HelixFinder::findAllHelicesGrow
	virtual ~AlignedHelix();

	string dotBracket(unsigned int alignmentLength);
	string dotBracket(unsigned int alignmentLength, unsigned int start);

	/**
	 * methods to collect stats on a given helix
	 */
	double pairedLikelihood(Tree * t);
	double unpairedLikelihood(Tree * t);
	double logLikeRatio(Tree * t);
	double canonicalBP();
	double covariance();
	double conservation();

	///////////////////////////////////////////////
	double singleLikelihood();
    double stackLikelihood();
    double logLikeRatioShape();
    static bool no_norm;
	///////////////////////////////////////////////

	bool operator< (const AlignedHelix & rhs) const;
	/**
	 * returns the number of bps in the helix
	 */
	int length();

	/**
	 * tries to insert the given helix
	 * If the helix aligns to exactly the same set of bps as this helix, add this sequence
	 * to the list of sequences in which this alignedHelix appears and return true.
	 * If not, do nothing and return false.
	 */
	bool insert(const Helix & helix, int seqIndex, vector<int> & seq2AlignmentMap);

	/**
	 * returns whether the helix is left of, right of, or overlapping the specified boundary index 
	 * where the boundary represents the position where two alignments were concatenated.
	 *  return: LEFT_OF_BOUNDARY, RIGHT_OF_BOUNDARY, OVERLAPPING_BOUNDARY, or INVALID_HELIX 
	 *  				(when the boundary is within the base pairs of the helix)
	 * by jeffp
	 */
	HelixPosition positionFromBoundary(int boundaryIndex);

	/**
	 * returns a string with the positions of the bps, separated by commas
	 * example:
	 * 12:32,11:31,10:29
	 */
	string bpsString();
	string bpsString(unsigned int start);

	/**
	 * vector containing pairs of integers representing the position of each bp of the helix.
	 * First in pair is 5' position
	 * Second in pair is 3' position
	 * Positions are with respect to the alignment.
	 * Bps in this vector are ordered outer to inner.
	 *
	 */
	vector<pair<int, int> > bps; //could be a linked list instead... not sure it would make a difference


private:
	Alignment* alignment;

};

#endif /* ALIGNEDHELIX_H_ */
