/*
 * Copyright: Nicholas P Wiebe (2008-2009) and Irmtraud M Meyer (2008-2009)
 * License: licensed under the GNU General Public License version 3 (GPLv3)
 * Alignment.cpp
 */

#include "Alignment.h"
#include "ShuffledAlignment.h"
#include "Tree.h"
#include "Utilities.h"
#include "LikelihoodCache.h"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <fstream>
#include <iostream>
#include <sstream>
#include <stack>
#include <string>
#include <stdexcept>
////////////////////
#include <unistd.h>
#include "models.h"
#include <sys/stat.h>
#include <sys/types.h>

#include "phy_files/utils.h"
#include "phy_files/utilsLinAlg.h"
#include "phy_files/GrammarIO.h"
#include "phy_files/GrammarWrap.h"
#include "phy_files/EmitWrapIO.h"

#include "phy_files/PhyDef.h"
#include "phy_files/DiscreteFactorGraph.h"
#include "phy_files/DfgIO.h"
//#include "phy_files/ama.h"
#include "phy_files/TRCTMCTree.h"
#include "phy_files/PhyIO.h"


/////////////////////////////////////
double Alignment::multiplier_shape = 1.0;
double Alignment::factor_fels = 1.0;
double Alignment::factor_shape = 1.0;
string Alignment::path_model = "";
bool Alignment::realign_cor = true;
string Alignment::realign_shapename = "";
/////////////////////////////////////

const string Alignment::TCOFFEE_LOC = "t_coffee";

int Alignment::minStemLength = 3;
int Alignment::minLoopLength = 2;

unsigned int refSeqNum=0; //reference sequence number from the alignment file is initially set to 0.

Alignment::Alignment() {
	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}

Alignment::Alignment(vector<string> names, vector<string> alignment, string codon){
	//copy names and alignments
	seqNames = names;
	alignedSeqs = alignment;

	codonAnnotation = codon;
	size_t match = codonAnnotation.find_first_of("123");
	annotated = (match != string::npos);

	this->initializeAlignmentMap();
	this->length = alignedSeqs[0].length();

	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}

//////////////////////////////////////////////////////////////
Alignment::Alignment(vector<string> names, vector<string> alignment, string codon, string shapefile, bool cor){
    //copy names and alignments
    seqNames = names;
    alignedSeqs = alignment;

    codonAnnotation = codon;
    size_t match = codonAnnotation.find_first_of("123");
    annotated = (match != string::npos);

    this->initializeAlignmentMap();
    this->length = alignedSeqs[0].length();

    consensusBP = NULL;
    seqCons = NULL;
    hasCache = false;

    this->readShape(shapefile, seq2AlignmentMap); //seq2AlignmentMap is a vector that has the mapped sequence positions.

    std::vector<string>::iterator it = std::find(seqs.begin(), seqs.end(), refSeq);
    refSeqID = std::distance(seqs.begin(), it);

    binValueCut = binValue;
    shapePosCut = shapePos;
    refSeqCut = refSeq;

//    makeCut(shapePos,binValue,refSeq);
    this->calculateDFGmodels(path_model, shapePosCut, binValueCut, refSeqCut);
//    corCut(singe_likelihoods,stack_mat,shapePos,binValue,refSeq);
    this->upperMatrixCopy(stack_mat);

}
//////////////////////////////////////////////////////////////


Alignment::Alignment(string & filename) {
	readAlignment(filename);

	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}


Alignment::Alignment(string & filename, string & annotation_file)
{
	readAlignment(filename);

	ifstream codons(annotation_file.c_str());
	string line;
	string seq;

	// Read in codon annotation
	codonAnnotation.clear();
	if(!codons.is_open()){
		cerr << "Error: cannot open codon annotation file " << annotation_file << endl;
		exit(-1);
	}
	while(codons.good()){
		getline(codons, line);
		Utilities::trimSpaces(line);
		//skip blank lines
		if(line.length() == 0){
			continue;
		}
		if(line.at(0) == '>'){
			// Ignore header line
		}
		else{
			seq = seq + line;
		}
	}

	// Check codon annotation validity
	if (seq.length() == 0) {
		cerr << "Error: No valid data found in " << annotation_file << endl;
		exit(-1);
	}
	size_t match = seq.find_first_of("0123");
	if (match != string::npos){
		// Check validity of codon annotation
		if(seq.length() != this->length) {
			cerr << "Error: Codon annotation length does not match that of first sequence" << endl;
			exit(-1);
		}
		size_t match = seq.find_first_not_of("0123");
		if (match != string::npos) {
			cerr << "Error: Invalid codon annotation '" << seq[match] << "' at position " << match << endl;
			exit(-1);
		}
		codonAnnotation = seq;
	} else {
		cerr << "Error: No valid codon annotation found in " << annotation_file << endl;
		exit(-1);
	}

	match = codonAnnotation.find_first_of("123");
	annotated = (match != string::npos) ? true : false;

	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}

/* ALIGNMENT CONSTRUCTOR
 * shapeFilename - generated shape by binGenerator.R
 * cor - use correlation model for DFG likelihoods
 *
*/
Alignment::Alignment(string & filename, string & shapeFilename, bool & cor)
{
    readAlignment(filename);

    readShape(shapeFilename, seq2AlignmentMap); //seq2AlignmentMap is a vector that has the mapped sequence positions.

    consensusBP = NULL;
    seqCons = NULL;
    hasCache = false;

    std::vector<string>::iterator it = std::find(seqs.begin(), seqs.end(), refSeq);
    refSeqID = std::distance(seqs.begin(), it);

    binValueCut = binValue;
    shapePosCut = shapePos;
    refSeqCut = refSeq;

//    makeCut(shapePos,binValue,refSeq);
    calculateDFGmodels(path_model, shapePosCut, binValueCut, refSeqCut);
//    corCut(singe_likelihoods,stack_mat,shapePos,binValue,refSeq);
    upperMatrixCopy(stack_mat);
}


/*
 * GENERATE DATA FOR DFG PROBABILITIES
 * Remove No Data objects, bcs they gives them priority over other no data.
 * Define values without data and generate positions, sequence,values which
 * are not equal to -1
*/
void Alignment::makeCut(vector<int> shapePos, vector<int> binValue, string refSeq) {
//    cout << "MakeCut exists" << endl;

    vector<int> binValueIndex;
    auto it = binValue.begin();
    while ((it= std::find_if(it, binValue.end(), [](int x){return x != -1; })) != binValue.end())
    {
        binValueIndex.push_back(std::distance(binValue.begin(), it));
        it++;
    }

    binValueCut.reserve(binValueIndex.size());
    shapePosCut.reserve(binValueIndex.size());
    refSeqCut = "";
    for(auto it = std::begin(binValueIndex); it != std::end(binValueIndex); ++it) {
        binValueCut.push_back(binValue[*it]);
        shapePosCut.push_back(shapePos[*it]);
        refSeqCut += refSeq[*it];
    }

    assert(binValueCut.size() == shapePosCut.size()); //sanity check
    assert(binValueCut.size() == refSeqCut.length()); //sanity check

//    cout << "NOT -1 index" << endl;
//    for (auto i = binValueIndex.begin(); i != binValueIndex.end(); ++i)
//        std::cout << *i << '\t';
//
//    cout << "\nbinValueCut" << endl;
//    for (auto i = binValueCut.begin(); i != binValueCut.end(); ++i)
//        std::cout << *i << '\t';
//
//    cout << "\nshapePosCut" << endl;
//    for (auto i = shapePosCut.begin(); i != shapePosCut.end(); ++i)
//        std::cout << *i << '\t';
//
//    cout << "\nrefSeqCut" << endl;
//    cout << refSeqCut << endl;

}


void Alignment::calculateDFGmodels(string path_model,vector<int> shapePosCut,
                              vector<int> binValueCut, string refSeqCut){

    phy::DfgInfo dfgInfo_single = phy::readDfgInfo(path_model + "/SingleStateMaps.txt",
                                                   path_model + "/SingleFactorPotentials.txt",
                                                   path_model + "/SingleVariables.txt",
                                                   path_model + "/SingleFactorGraph.txt");
    phy::DfgInfo dfgInfo_stack = phy::readDfgInfo(path_model + "/StackStateMaps.txt",
                                                  path_model + "/StackFactorPotentials.txt",
                                                  path_model + "/StackVariables.txt",
                                                  path_model + "/StackFactorGraph.txt");
    phy::DfgInfo dfgInfo_pair = phy::readDfgInfo(path_model + "/PairStateMaps.txt",
                                                  path_model + "/PairFactorPotentials.txt",
                                                  path_model + "/PairVariables.txt",
                                                  path_model + "/PairFactorGraph.txt");

    vector<phy::SeqToVarSymbol> stvVec_single = phy::readSeqToVarSymbols(path_model + "/SingleSeqToVarSymbol.txt");
    vector<phy::SeqToVarSymbol> stvVec_stack = phy::readSeqToVarSymbols(path_model + "/StackSeqToVarSymbol.txt");
    vector<phy::SeqToVarSymbol> stvVec_pair = phy::readSeqToVarSymbols(path_model + "/PairSeqToVarSymbol.txt");
    bool DFG_correlation;
    // DEFINE MODEL TYPE BASED ON DFG PROPERTIES FOR SHAPE DATA
    // USED IN GenAma() FOR SPECIAL FORMAT OF CODING SHAPE DATA
    for (int i = 0; i < stvVec_single.size(); ++i) {
        if (stvVec_single.at(i).varName == "d" && stvVec_single.at(i).seqName == "shape"){
            if (stvVec_single.at(i).size == 1 && stvVec_single.at(i).symbolOffset == 1 &&
                stvVec_single.at(i).rSize == 1 && stvVec_single.at(i).indexOffset == 0 &&
                stvVec_single.at(i).rIndexOffset == 0 ){
                DFG_correlation = true;
            }
            else if (stvVec_single.at(i).size == 2 && stvVec_single.at(i).symbolOffset == 2 &&
                     stvVec_single.at(i).rSize == 2 && stvVec_single.at(i).indexOffset == 0 &&
                     stvVec_single.at(i).rIndexOffset == 0 ){
                DFG_correlation = false;
            }
            break;
        }
    }
    //GENERATE STRING OF AMA FILE
    if(DFG_correlation){
        genAma(shapePos,binValue,refSeq,DFG_correlation);
    }else{
        genAma(shapePos,binValue,refSeq,DFG_correlation);
    }

    //UPDATED READING OF AMA FILES
    //DO NOT REQUIRE WRITING TO RANDOM FOLDER
    vector<phy::ama> ama_vector_from_string = phy::readAmaString(ama_str);
    vector<phy::SeqData> seqV_string = phy::amaToSeqData(ama_vector_from_string);
    //CALCULATING SINGLE PROBABILITIES
    phy::xvector_t result_single;
    result_single = phy::calcLikelihoodVector(seqV_string[0], dfgInfo_single, stvVec_single);

    singe_likelihoods.clear();
    singe_likelihoods.reserve(result_single.size());
    for (unsigned i = 0; i < result_single.size(); i++) {
        singe_likelihoods.push_back(phy::toNumber(result_single[i]));
    }
//    cout << "SIZE OF singe_likelihoods IS:\t"<< singe_likelihoods.size() << endl;
    //CALCULATING STACK PROBABILITIES
    phy::xmatrix_t result_stack;
    result_stack = calcLikelihoodMatrix(seqV_string[0], dfgInfo_stack, stvVec_stack);
    stack_mat = phy::toNumber(result_stack);

    //CALCULATING PAIR PROBABILITIES
    phy::xmatrix_t result_pair;
    result_pair = calcLikelihoodMatrix(seqV_string[0], dfgInfo_pair, stvVec_pair);
    pair_mat = phy::toNumber(result_pair);

}



/*
 * WITH CORRELATION DELETE BINS WITH -1 INSIDE
 * singe_likelihoods AND stack_mat
*/
void Alignment::upperMatrixCopy(boost::numeric::ublas::matrix<double> stack_mat){
//    cout << "upperMatrixCopy exists" << endl;

    for (unsigned i = 0; i < stack_mat.size1(); ++i) {
        for (unsigned j = 0; j < stack_mat.size2(); ++j) {
            stack_mat(i, j) = stack_mat(j, i);
        }
    }
}

/*
 * WITH CORRELATION DELETE BINS WITH -1 INSIDE
 * singe_likelihoods AND stack_mat
*/
void Alignment::corCut(vector<double> singe_likelihoods,boost::numeric::ublas::matrix<double> stack_mat,
                       vector<int> shapePos, vector<int> binValue, string refSeq) {

//    cout << "corCut exists" << endl;

    vector<int> binValueIndex;
    auto it = binValue.begin();
    while ((it= std::find_if(it, binValue.end(), [](int x){return x != -1; })) != binValue.end())
    {
        binValueIndex.push_back(std::distance(binValue.begin(), it));
        it++;
    }

    vector<double> singe_likelihoods_tmp;
    singe_likelihoods_tmp.reserve(binValueIndex.size());
    for(auto it = std::begin(binValueIndex); it != std::end(binValueIndex); ++it) {
        singe_likelihoods_tmp.push_back(singe_likelihoods[*it]);
    }
    singe_likelihoods.clear();
    singe_likelihoods = singe_likelihoods_tmp;

    numeric::ublas::matrix<double> stack_mat_tmp(binValueIndex.size(),binValueIndex.size());
    for (unsigned i = 0; i < binValueIndex.size(); ++i) {
        for (unsigned j = 0; j < binValueIndex.size(); ++j) {
            stack_mat_tmp(i, j) = stack_mat(binValueIndex[i], binValueIndex[j]);
        }
    }
    stack_mat.clear();
    stack_mat = stack_mat_tmp;
}



/*
 * READ DFG MODEL AND CREATE LIKELIHOODS VECTOR AND MATRIX FOR SINGLE AND STACK
 * AFTER READ DELETE "RANDOM" FOLDER
*/
/*void Alignment::DfgLikelihoods(string name_rand) {
//    cout << "DfgLikelihoods exists" << endl;

    phy::DfgInfo dfgInfo_single = phy::readDfgInfo(name_rand + "/SingleStateMaps.txt",name_rand + "/SingleFactorPotentials.txt",name_rand + "/SingleVariables.txt",name_rand + "/SingleFactorGraph.txt");
    phy::DfgInfo dfgInfo_stack = phy::readDfgInfo(name_rand + "/StackStateMaps.txt",name_rand + "/StackFactorPotentials.txt",name_rand + "/StackVariables.txt",name_rand + "/StackFactorGraph.txt");

    vector<phy::SeqToVarSymbol> stvVec_single = phy::readSeqToVarSymbols(name_rand + "/SingleSeqToVarSymbol.txt");
    vector<phy::SeqToVarSymbol> stvVec_stack = phy::readSeqToVarSymbols(name_rand + "/StackSeqToVarSymbol.txt");

    //read seq data via ama file
    vector<phy::ama> amaV = phy::readAmaFile(name_rand + "/ama.txt" );
    vector<phy::SeqData> seqV = phy::amaToSeqData(amaV);

    //delete folder
    remove((name_rand + "/ama.txt").c_str());
    remove((name_rand + "/SingleStateMaps.txt").c_str());
    remove((name_rand + "/SingleFactorGraph.txt").c_str());
    remove((name_rand + "/SingleFactorPotentials.txt").c_str());
    remove((name_rand + "/SingleSeqToVarSymbol.txt").c_str());
    remove((name_rand + "/SingleVariables.txt").c_str());
    remove((name_rand + "/StackFactorGraph.txt").c_str());
    remove((name_rand + "/StackFactorPotentials.txt").c_str());
    remove((name_rand + "/StackSeqToVarSymbol.txt").c_str());
    remove((name_rand + "/StackStateMaps.txt").c_str());
    remove((name_rand + "/StackVariables.txt").c_str());
    remove(name_rand.c_str());

    phy::xvector_t result_single;
    result_single = phy::calcLikelihoodVector(seqV[0], dfgInfo_single, stvVec_single);
    singe_likelihoods.clear();
    singe_likelihoods.reserve(result_single.size());
    for (unsigned i = 0; i < result_single.size(); i++) {
        singe_likelihoods.push_back(phy::toNumber(result_single[i]));
    }

//    cout << "SIZE OF singe_likelihoods IS:\t"<< singe_likelihoods.size() << endl;

    phy::xmatrix_t result_stack;
    result_stack = calcLikelihoodMatrix(seqV[0], dfgInfo_stack, stvVec_stack);
    stack_mat = phy::toNumber(result_stack);

//    cout << "LIKELIHOODS DONE" << endl;

}*/

/*
 * WRITE AMA AND DFG PARAMETERS TO "RANDOM" FOLDER
 * ama_str ==> ama.txt
 * They wil have correlation if cor bool is TRUE
 * SINGLE:
 * SingleStateMaps ==> "SingleStateMaps.txt"
 * SingleFactorPotentials ==> "SingleFactorPotentials.txt"
 * SingleVariables ==> "SingleVariables.txt"
 * SingleFactorGraph ==> "SingleFactorGraph.txt"
 * SingleSeqToVarSymbol ==> "SingleSeqToVarSymbol.txt"
 *
 * STACK:
 * StackStateMaps ==> "StackStateMaps.txt"
 * StackFactorPotentials ==> "StackFactorPotentials.txt"
 * StackVariables ==> "StackVariables.txt"
 * StackFactorGraph ==> "StackFactorGraph.txt"
 * StackSeqToVarSymbol ==> "StackSeqToVarSymbol.txt"
 *
*/
/*void Alignment::writeFilesAmaDfg(string name_rand, bool cor) {
//    cout << "writeFilesAmaDfg exists" << endl;

    mkdir(name_rand.c_str(), 0777);

    ofstream os;
    os.open(name_rand + "/ama.txt");
    os << ama_str;
    os.close();

    os.open(name_rand + "/SingleStateMaps.txt");
    if(cor){
        os << models::SingleStateMapsCor;
    }else{
        os << models::SingleStateMapsNoCor;
    }
    os.close();

    os.open(name_rand + "/SingleFactorPotentials.txt");
    if(cor){
        os << models::SingleFactorPotentialsCor;
    }else{
        os << models::SingleFactorPotentialsNoCor;
    }
    os.close();

    os.open(name_rand + "/SingleVariables.txt");
    if(cor){
        os << models::SingleVariablesCor;
    }else{
        os << models::SingleVariablesNoCor;
    }
    os.close();

    os.open(name_rand + "/SingleFactorGraph.txt");
    if(cor){
        os << models::SingleFactorGraphCor;
    }else{
        os << models::SingleFactorGraphNoCor;
    }
    os.close();

    os.open(name_rand + "/SingleSeqToVarSymbol.txt");
    if(cor){
        os << models::SingleSeqToVarSymbolCor;
    }else{
        os << models::SingleSeqToVarSymbolNoCor;
    }
    os.close();

    os.open(name_rand + "/StackStateMaps.txt");
    if(cor){
        os << models::StackStateMapsCor;
    }else{
        os << models::StackStateMapsNoCor;
    }
    os.close();

    os.open(name_rand + "/StackFactorPotentials.txt");
    if(cor){
        os << models::StackFactorPotentialsCor;
    }else{
        os << models::StackFactorPotentialsNoCor;
    }
    os.close();

    os.open(name_rand + "/StackVariables.txt");
    if(cor){
        os << models::StackVariablesCor;
    }else{
        os << models::StackVariablesNoCor;
    }
    os.close();

    os.open(name_rand + "/StackFactorGraph.txt");
    if(cor){
        os << models::StackFactorGraphCor;
    }else{
        os << models::StackFactorGraphNoCor;
    }
    os.close();

    os.open(name_rand + "/StackSeqToVarSymbol.txt");
    if(cor){
        os << models::StackSeqToVarSymbolCor;
    }else{
        os << models::StackSeqToVarSymbolNoCor;
    }
    os.close();

}*/

/*
 * GENERATE AMA FOR DFG PROBABILITIES WITH CORRELATION AND NO CORRELATION
 * Take cut data(reducing computing time)
 * Generate MAF style file for PHY recognition of it annotated
 * multiple alignment format (modeled over the maf format:
 * http://genome.ucsc.edu/FAQ/FAQformat#format5
 * (or originally: http://eyebrowse.cit.nih.gov/genome/goldenPath/help/maf.html.)
*/
void Alignment::genAma(vector<int> shapePosCut, vector<int> binValueCut, string refSeqCut, bool cor) {
//    cout << "genAma exists" << endl;

    ama_str = "a score=0.1 ENTRY=RNA\n";
    ama_str = ama_str + "s\tdna\t0\t" + to_string(refSeqCut.length()) + "\t+\t" + to_string(refSeqCut.length()) + "\t" + refSeqCut + "\n" ;
    ama_str = ama_str + "s\tshape\t0\t" + to_string(refSeqCut.length()) + "\t+\t" + to_string(refSeqCut.length()) + "\t";

    if (cor){
        for (auto i = binValueCut.begin(); i != binValueCut.end(); ++i){
            if(*i == -1){
                ama_str = ama_str + "*";
            }else{
                ama_str = ama_str + to_string(*i);
            }
        }
    }else{
        for (auto i = binValueCut.begin(); i != binValueCut.end(); ++i){
            if(*i == -1){
                ama_str = ama_str + "**";
            }else{
                ama_str = ama_str + "0" + to_string(*i);
            }
        }
    }
    ama_str = ama_str + "\n";
//    cout << ama_str << endl;
}


/*
 * GENERATE RANDOM NAME FOR TEMP FOLDER
 */
//void Alignment::genRandomName(const int len) {
//    string tmp_s;
//    static const char alphanum[] =
//            "0123456789"
//            "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
//            "abcdefghijklmnopqrstuvwxyz";
//
//    srand( (unsigned) time(NULL) * getpid());
//    tmp_s.reserve(len);
//
//    for (int i = 0; i < len; ++i)
//        tmp_s += alphanum[rand() % (sizeof(alphanum) - 1)];
//
//    name_rand = tmp_s;
//}

/*seq2Alignment has MSA mapped nucleotide positions of the reference sequence
 *mapping the ungapped nucleotide positions from the shape file to the MSA
 *finds the reference sequence position in the MSA, if not found, throws an error and forced exit
 *reserves memory space for shape postions vector and bin values vector
 */
void Alignment::fillShape(vector <int> tempShapePos, vector <int> tempBinValue, string refSeq)
{

    int count = 0;
    int refSeqFound = false;

    //adding gaps to the raw reference sequence by mapping it to the MSA file to identify the reference sequence form the MSA
    vector<basic_string<char>>::iterator iter;
    for(iter = seqs.begin(); iter != seqs.end(); ++iter){
        if(refSeq == *iter){
            refSeqNum = count;
            refSeqFound = true;
            break;
        }
        else{
            count++;
        }
    }
    //cout << "reference sequence number in MSA is " << refSeqNum << endl;

    if(!refSeqFound){
        cerr << "Reference sequence not found in your input MSA file. " << endl;
        exit(-1);
    }

    shapePos.reserve(seq2AlignmentMap[refSeqNum]->size());
    binValue.reserve(seq2AlignmentMap[refSeqNum]->size());

    //maps the MSA mapped nucleotide positions of the reference sequence
    for (unsigned int i=0; i < shapePos.capacity();i++){
        shapePos.push_back(seq2AlignmentMap[refSeqNum]->at(i));
        binValue.push_back(tempBinValue[i]);
        assert(shapePos.size() == binValue.size());
    }
}

/*reads the input SHAPE bin file*/
void Alignment::readShape(string& filename, vector<vector<int>* >& seq2AlignmentMap)
{
    ifstream shapeFile(filename.c_str());
    string line;
    int count=0; //counts the number of non blank lines leaving out the header line
    int behind;
    int ahead;
    int space;
    string refSeqNucl;
    refSeq=""; //has the reference sequence
    refSeqStore="";

    //stores the MSA mapped nucleotide positions of the reference sequence
    vector <int> tempShapePos;

    //stores the bin values corresponding to tempShapePos
    vector <int> tempBinValue;

    if(!shapeFile.is_open()){ //sanity check
        cerr << "Error: cannot open shape file " << filename << endl;
        exit(-1);
    }

    shapeFile.ignore(10000, '\n'); //ignores the header line of the shape input file

    while(shapeFile.good()){
        getline(shapeFile, line);
        Utilities::trimSpaces(line);

        //skip blank lines
        if(line.length() == 0){
            continue;
        }
        count++;

        space = line.find_first_of("\t");
        behind =std::atoi(line.substr(0,space).c_str());
        tempShapePos.push_back(behind);

        ahead = std::atoi(line.substr(space+1).c_str());

        //0, 1, 2, 3, 4, 5, -1 are allowed bin values. -1: missing shape data
        if(ahead == 0 || ahead == 1 || ahead == 2 || ahead == 3 || ahead == 4 || ahead == 5 || ahead == -1){
            tempBinValue.push_back(ahead);
        }
        else{
            cerr << "Error: Bin number other than 0-5 detected in your input shape file.\n";
            exit(-1);
        }

        space = line.find_last_of("\t");
        refSeqNucl = line.substr(space+1);
        refSeq += refSeqNucl; //stores the reference sequence (adding nucleotide by nucleotide)
    }

    refSeqStore = refSeq;
//    cout << "refSeq from input SHAPE file is \n" << refSeq << endl;

    fillShape(tempShapePos, tempBinValue, refSeq);
    shapeFile.close();

    assert(seq2AlignmentMap[refSeqNum]->size() == shapePos.size()); //sanity check

}


Alignment::Alignment(Alignment* a, Alignment* b) {

	assert(a->seqNames.size() == b->seqNames.size());

	for (unsigned int i = 0 ; i < a->seqNames.size() ; i++) {
		seqNames = a->seqNames;
		alignedSeqs.push_back(a->alignedSeqs[i] + b->alignedSeqs[i]);
	}

	// Combine codon annotations
	codonAnnotation = a->codonAnnotation;
	codonAnnotation.insert(codonAnnotation.end(), b->codonAnnotation.begin(), b->codonAnnotation.end());
	annotated = a->annotated || b->annotated;

	this->initializeAlignmentMap();
	this->length = alignedSeqs[0].length();
	consensusBP = NULL;
	seqCons = NULL;
	hasCache = false;
}

//extract unaligned seqs and make map of unaligned seq positions to aligned seq positions
void Alignment::initializeAlignmentMap() {
	vector<int>* temp;
	vector<int>* temp2;
	string seq;

	unsigned int alignedSeqLength = alignedSeqs[0].length();

	for (unsigned int i = 0; i < alignedSeqs.size(); i++){

		assert(alignedSeqLength == alignedSeqs[i].length()); //sanity check
		temp = new vector<int>();
		temp2 = new vector<int>();
		seq = "";

		for (unsigned int j = 0; j < alignedSeqLength; j++){
			if (alignedSeqs[i].at(j) != '-'){
				temp->push_back(j);
				seq += alignedSeqs[i].at(j);
				temp2->push_back(temp->size() - 1);
			}
			else{
				temp2->push_back(-1);
			}
		}
		seqs.push_back(seq);
		seq2AlignmentMap.push_back(temp);
		alignment2SeqMap.push_back(temp2);
	}
}


/*
 * This creates the likelihood cache object
 * You must call this method on the Alignment object if you wish to
 * calculate log likelihood values.
 */
void Alignment::initializeCache() {
	hasCache = true;
	likelihoods = new LikelihoodCache(this->length);
}

Alignment::~Alignment() {
	size_t i;
	for (i = 0; i < seq2AlignmentMap.size(); i++){
		delete seq2AlignmentMap[i];
	}
	seq2AlignmentMap.clear();

	for (i = 0; i < alignment2SeqMap.size(); i++){
		delete alignment2SeqMap[i];
	}
	alignment2SeqMap.clear();

	if (hasCache) {
		delete likelihoods;
		likelihoods = NULL;
	}

	delete consensusBP;
	consensusBP = NULL;

	delete seqCons;
	seqCons = NULL;
}

string Alignment::seqString()
{
	string out = "";
	for(unsigned int i = 0; i < seqs.size(); i++){
		out += seqs[i];
		out += "\n";
	}
	return out;
}

string Alignment::columnString(int pos)
{
	string out = "";
	for(unsigned int i = 0; i < alignedSeqs.size(); i++){
		//cout << *(alignedSeqs[i]) << endl;
		out += alignedSeqs[i].at(pos);
	}
	return out;
}


string Alignment::alignmentString()
{
	// cout << "Alignment Size: " << alignedSeqs.size() << endl;
	string out = "";
	for(unsigned int i = 0; i < alignedSeqs.size(); i++){
		//cout << *(alignedSeqs[i]) << endl;
		out += alignedSeqs[i];
		out += "\n";
	}
	return out;
}

string Alignment::codonString()
{
	return codonAnnotation;
}

void Alignment::readAlignment(string & filename)
{
	ifstream fastaAln(filename.c_str());
	string line;
	string seq = "";
	string name = "";
	if(!fastaAln.is_open()){
		cerr << "Error: cannot open alignment file " << filename << endl;
		exit(-1);
	}
	while(fastaAln.good()){
		getline(fastaAln, line);
		Utilities::trimSpaces(line);
		//skip blank lines
		if(line.length() == 0){
			continue;
		}
		if(line.at(0) == '>'){
			// Write previous entry
			if(seq != ""){
				//convert to upper case, and convert Ts to Us
				for(unsigned int i = 0; i < seq.length(); i++){
					seq[i] = toupper(seq[i]);
					if(seq[i] == 'T'){
						seq[i] = 'U';
					}
					if(seq[i] == '.') {
						seq[i] = '-';
					}
				}
				alignedSeqs.push_back(seq);
			}
			// Start new entry
			seqNames.push_back(line.substr(1));
			seq = "";
		}
		else{
			seq = seq + line;
		}
	}

	// Process last line
	assert(seq.length() > 0);

	size_t match = seq.find_first_of("0123");
	if (match != string::npos){
		//last sequence contains the codon annotation

		// Check validity of codon annotation
		if(seq.length() != alignedSeqs[0].size()) {
			cerr << "Error: Codon annotation length does not match that of first sequence" << endl;
			exit(-1);
		}
		size_t match = seq.find_first_not_of("0123");
		if (match != string::npos) {
			cerr << "Error: Invalid codon annotation '" << seq[match] << "' at position " << match << endl;
			exit(-1);
		}

		seqNames.pop_back(); // remove last species name (corresponding to codon annotation)
		codonAnnotation = seq;

		match = seq.find_first_of("123");
		annotated = (match != string::npos) ? true : false;
	}
	else{
		annotated = false;
		//convert to upper case, and convert Ts to Us
		for(unsigned int i = 0; i < seq.length(); i++){
			seq[i] = toupper(seq[i]);
			if(seq[i] == 'T'){
				seq[i] = 'U';
			}
			codonAnnotation += "0";
		}
		alignedSeqs.push_back(seq);
	}

	fastaAln.close();

	this->initializeAlignmentMap();
	this->length = alignedSeqs[0].length();
}


/*
 * The following three functions are used to retrieve likelihoods from the alignment
 * Precondition: you must first call initializeCache() before calling any of these
 * methods. This saves space for alignments which are not used to calculate likelihoods,
 * as they do not require cache memory
 */
/*double Alignment::logLikelihood(int pos5, int pos3, Tree * tree){

    auto it = find(this->shapePosCut.begin(), this->shapePosCut.end(), pos5);
    int index_pos5;
    if (it != this->shapePosCut.end()){
        index_pos5 = it - this->shapePosCut.begin();
    }else {
        index_pos5 = -1;
    }

    it = find(this->shapePosCut.begin(), this->shapePosCut.end(), pos3);
    int index_pos3;
    if (it != this->shapePosCut.end()){
        index_pos3 = it - this->shapePosCut.begin();
    }else {
        index_pos3 = -1;
    }

//    cout << "\npos5:\t" << pos5 << "\tindex:\t" << index_pos5 << "\t";
//    if (index_pos5 == -1){
//        cout << "NO VALUE" << endl;
//    }else{
//        cout << "singe_likelihoods:\t"<< this->singe_likelihoods[index_pos5] << endl;
//    }

//    cout << "pos3:\t" << pos3 << "\tindex:\t" << index_pos3 << "\t";
//    if (index_pos3 == -1){
//        cout << "NO VALUE" << endl;
//    }else{
//        cout << "singe_likelihoods:\t"<< this->singe_likelihoods[index_pos3]<< endl;
//    }

    if(index_pos3 != -1 &&  index_pos5 != -1){
        if(this->stack_mat(index_pos5,index_pos3) == 0.0){
            return this->logPairedLikelihood(pos5, pos3, tree)
                   - this->logUnpairedLikelihood(pos5, tree)
                   - this->logUnpairedLikelihood(pos3, tree);
        }else{
//            long double LFP = tree->calcLikelihoodFelsPaired(*this, pos5, pos3);
//            long double LFUP5 = tree->calcLikelihoodFelsUnpaired(*this, pos5);
//            long double LFUP3 = tree->calcLikelihoodFelsUnpaired(*this, pos3);
//
//            long double LST = this->stack_mat(index_pos5,index_pos3);
//            long double LS5 = this->singe_likelihoods[index_pos5];
//            long double LS3 = this->singe_likelihoods[index_pos3];
//
//            long double LLFP =this->logPairedLikelihood(pos5, pos3, tree);
//            long double LLFUP5 = this->logUnpairedLikelihood(pos5, tree);
//            long double LLFUP3 = this->logUnpairedLikelihood(pos3, tree);
//
//            long double LLST = log(this->stack_mat(index_pos5,index_pos3))/log(2.0);
//            long double LLS5 = log(this->singe_likelihoods[index_pos5])/log(2.0);
//            long double LLS3 = log(this->singe_likelihoods[index_pos3])/log(2.0);

            return this->logPairedLikelihood(pos5, pos3, tree)
                   - this->logUnpairedLikelihood(pos5, tree)
                   - this->logUnpairedLikelihood(pos3, tree)
                   + this->logStackLikelihood(pos5,pos3,index_pos5,index_pos3)
                   - this->logSingleLikelihood(pos5, index_pos5)
                   - this->logSingleLikelihood(pos3, index_pos3);

//            return (LLFP-LLFUP5-LLFUP3+multiplier_shape*LLST-multiplier_shape*LLS5-multiplier_shape*LLS3);
        }
    }else{
        return this->logPairedLikelihood(pos5, pos3, tree)
               - this->logUnpairedLikelihood(pos5, tree)
               - this->logUnpairedLikelihood(pos3, tree);
    }

}*/

//double Alignment::logLikelihood(int pos5, int pos3, Tree * tree){
//	return this->logPairedLikelihood(pos5, pos3, tree)
//			- this->logUnpairedLikelihood(pos5, tree)
//			- this->logUnpairedLikelihood(pos3, tree);
//}

double Alignment::logLikelihood(int pos5, int pos3, Tree * tree) {

    auto it = find(this->shapePosCut.begin(), this->shapePosCut.end(), pos5);
    int index_pos5;
    if (it != this->shapePosCut.end()) {
        index_pos5 = it - this->shapePosCut.begin();
    } else {
        index_pos5 = -1;
    }

    it = find(this->shapePosCut.begin(), this->shapePosCut.end(), pos3);
    int index_pos3;
    if (it != this->shapePosCut.end()) {
        index_pos3 = it - this->shapePosCut.begin();
    } else {
        index_pos3 = -1;
    }

    if(index_pos3 != -1 &&  index_pos5 != -1) {
        if (this->stack_mat(index_pos5, index_pos3) == 0.0) {
            return factor_fels * this->logPairedLikelihood(pos5, pos3, tree)
                   - factor_fels * this->logUnpairedLikelihood(pos5, tree)
                   - factor_fels * this->logUnpairedLikelihood(pos3, tree);
        } else {
            return factor_fels * this->logPairedLikelihood(pos5, pos3, tree)
                   - factor_fels * this->logUnpairedLikelihood(pos5, tree)
                   - factor_fels * this->logUnpairedLikelihood(pos3, tree)
                   + factor_shape * this->logStackLikelihood(pos5, pos3, index_pos5, index_pos3)
                   - factor_shape * this->logSingleLikelihood(pos5, index_pos5)
                   - factor_shape * this->logSingleLikelihood(pos3, index_pos3);
        }
    }
}

double Alignment::logLikelihoodPair(int pos5, int pos3, Tree * tree) {

    auto it = find(this->shapePosCut.begin(), this->shapePosCut.end(), pos5);
    int index_pos5;
    if (it != this->shapePosCut.end()) {
        index_pos5 = it - this->shapePosCut.begin();
    } else {
        index_pos5 = -1;
    }

    it = find(this->shapePosCut.begin(), this->shapePosCut.end(), pos3);
    int index_pos3;
    if (it != this->shapePosCut.end()) {
        index_pos3 = it - this->shapePosCut.begin();
    } else {
        index_pos3 = -1;
    }

    if(index_pos3 != -1 &&  index_pos5 != -1) {
        if (this->stack_mat(index_pos5, index_pos3) == 0.0) {
            return factor_fels * this->logPairedLikelihood(pos5, pos3, tree)
                   - factor_fels * this->logUnpairedLikelihood(pos5, tree)
                   - factor_fels * this->logUnpairedLikelihood(pos3, tree);
        } else {
            return factor_fels * this->logPairedLikelihood(pos5, pos3, tree)
                   - factor_fels * this->logUnpairedLikelihood(pos5, tree)
                   - factor_fels * this->logUnpairedLikelihood(pos3, tree)
                   + factor_shape * this->logPairLikelihood(pos5, pos3, index_pos5, index_pos3)
                   - factor_shape * this->logSingleLikelihood(pos5, index_pos5)
                   - factor_shape * this->logSingleLikelihood(pos3, index_pos3);
        }
    }
}


double Alignment::logPairedLikelihood(int pos5, int pos3, Tree * tree){
	if (hasCache && likelihoods->isCached(pos5, pos3)) {
		return likelihoods->getCached(pos5, pos3);
	}
	else {
		double likelihood = tree->calcLogFelsPaired(*this, pos5, pos3);
		if (hasCache) likelihoods->setCached(pos5, pos3, likelihood);
		return likelihood;
	}
}

double Alignment::logUnpairedLikelihood(int pos, Tree * tree){
	if (hasCache && likelihoods->isCached(pos)) {
		return likelihoods->getCached(pos);
	}
	else {
		double likelihood = tree->calcLogFelsUnpaired(*this, pos);
		if (hasCache) likelihoods->setCached(pos, likelihood);
		return likelihood;
	}
}

/////////////////////////////////////////////////////////

double Alignment::logStackLikelihood(int pos5, int pos3,int index_pos5,int index_pos3){
    double likelihood = log(this->stack_mat(index_pos5, index_pos3)) / log(2.0);
    return likelihood;
}

double Alignment::logPairLikelihood(int pos5, int pos3,int index_pos5,int index_pos3){
    double likelihood = log(this->pair_mat(index_pos5, index_pos3)) / log(2.0);
    return likelihood;
}

//double Alignment::logStackLikelihood(int pos5, int pos3,int index_pos5,int index_pos3){
//    if (hasCache && likelihoods->isCachedShape(pos5, pos3)) {
//        return likelihoods->getCachedShape(pos5, pos3);
//    }
//    else {
//        double likelihood = log(this->stack_mat(index_pos5, index_pos3)) / log(2.0);
//        return likelihood;
//    }
//}

double Alignment::logSingleLikelihood(int pos, int index_pos){
    double likelihood = log(this->singe_likelihoods[index_pos]) / log(2.0);
    return likelihood;
}

//double Alignment::logSingleLikelihood(int pos, int index_pos){
//    if (hasCache && likelihoods->isCachedShape(pos)) {
//        return likelihoods->getCachedShape(pos);
//    }
//    else {
//        double likelihood = log(this->singe_likelihoods[index_pos]) / log(2.0);
//        return likelihood;
//    }
//}

double Alignment::logLikelihoodShape(int pos5, int pos3,int index_pos5,int index_pos3){
    double likelihood = this->logStackLikelihood(pos5,pos3,index_pos5,index_pos3)
                            - this->logSingleLikelihood(pos5, index_pos5)
                            - this->logSingleLikelihood(pos3, index_pos3);
    return likelihood;
}

double Alignment::logLikelihoodShapePair(int pos5, int pos3,int index_pos5,int index_pos3){
    double likelihood = this->logPairLikelihood(pos5,pos3,index_pos5,index_pos3)
                        - this->logSingleLikelihood(pos5, index_pos5)
                        - this->logSingleLikelihood(pos3, index_pos3);
    return likelihood;
}

//double Alignment::logLikelihoodShape(int pos5, int pos3,int index_pos5,int index_pos3){
//
//    if (hasCache && likelihoods->isCachedShape(pos5, pos3)) {
//        return likelihoods->getCachedShape(pos5, pos3);
//    }
//    else {
//        double likelihood = this->logStackLikelihood(pos5,pos3,index_pos5,index_pos3)
//                            - this->logSingleLikelihood(pos5, index_pos5)
//                            - this->logSingleLikelihood(pos3, index_pos3);
//        return likelihood;
//    }
//}

Alignment* Alignment::realignTcoffeeShape(Tree * tree){
    //write fasta file:
    string fastaFilenameTemplate = "/tmp/fasta.XXXXXX";
    char fastaFilename[fastaFilenameTemplate.size()+1];
    strcpy(fastaFilename, fastaFilenameTemplate.c_str());

    int unique_fd;
    FILE * unique_file;

    if ((unique_fd = mkstemp(fastaFilename)) == -1 ){
        cerr << "Could not create temporary clustal file for shuffling\n";
        exit(-1);
    }

    if((unique_file = fdopen(unique_fd, "w")) == NULL){
        cerr << "Could not open temporary clustal file for shuffling\n";
        exit(-1);
    }

    fputs(FastaFormat(false).c_str(), unique_file);
    fclose(unique_file);

    //make a temp file for output:
    string outTemplate = "/tmp/realign_out.XXXXXX";
    char outFilename[outTemplate.size()+1];
    strcpy(outFilename, outTemplate.c_str());

    if ((unique_fd = mkstemp(outFilename)) == -1 ){
        cerr << "Could not create temporary out file for realignment\n";
        exit(-1);
    }

    if((unique_file = fdopen(unique_fd, "w")) == NULL){
        cerr << "Could not open temporary out file for realignment\n";
        exit(-1);
    }
    string temp_contents = "TEMP!\n"; // this will be replaced by TCOFFEE output
    fputs(temp_contents.c_str(), unique_file);
    fclose(unique_file);

    //make a temp file for output:
    string treeTemplate = "/tmp/tree_out.XXXXXX";
    char treeFilename[treeTemplate.size()+1];
    strcpy(treeFilename, treeTemplate.c_str());

    if ((unique_fd = mkstemp(treeFilename)) == -1 ){
        cerr << "Could not create temporary tree file for realignment\n";
        exit(-1);
    }

    if((unique_file = fdopen(unique_fd, "w")) == NULL){
        cerr << "Could not open temporary out file for realignment\n";
        exit(-1);
    }
    //write tree to file
    fputs(tree->newickString().c_str(), unique_file);
    fclose(unique_file);

//    cout << string(fastaFilename) << "\n" << string(outFilename) << "\n" << string(treeFilename) << endl;

    //run realignment:
    string command = TCOFFEE_LOC + " " + string(fastaFilename) + " -outfile " + string(outFilename) + " -usetree " + string(treeFilename) + " -output fasta -n_core 1 -outorder input -quiet";
    if(system(command.c_str()) != 0){
        cerr << "Error: t_coffee encountered an error\n";
        exit(-1);
    }

    ifstream realignFile(outFilename);

    if(!realignFile.is_open()){
        cerr << "Error: cannot open realignment fasta file " << string(outFilename) << endl;
        exit(-1);
    }

    string line;
    vector<string> realignment;
    vector<string> r_names;
    string seq;
    bool found_seq = false;

    while(realignFile.good()){
        getline(realignFile, line);
        Utilities::trimSpaces(line);
        //skip blank lines
        if(line.length() == 0){
            continue;
        }
        if(line.at(0) == '>'){
            if(found_seq){
                //cout << "Sequence Found: " << *seq << endl;
                realignment.push_back(seq);
            }
            line = line.substr(1);
            r_names.push_back(line);
            found_seq = true;
            seq = "";
        }
        else{
            seq = seq + line;
        }
    }

    realignFile.close();
    realignment.push_back(seq);

    assert(realignment.size() == r_names.size());
    assert(realignment.size() == alignedSeqs.size());

    for(unsigned int j = 0; j < r_names.size(); j++){
        assert(r_names[j].compare(seqNames[j]) == 0);
        assert(realignment[j].length() == realignment[0].length());
    }

    Alignment* resultAlignment = new Alignment(this->seqNames, realignment, this->codonAnnotation,
                                               this->realign_shapename, this->realign_cor);
//    Alignment(vector<string> names, vector<string> alignment, string codon, string shapefile, bool cor)
    resultAlignment->realignCodonAnnotation(this->alignedSeqs[0]);

    //clean up files
    command = "rm -f " + string(outFilename) + " " + string(fastaFilename) + " " + string(treeFilename) ;
    if(system(command.c_str()) != 0){
        cerr << "Error cleaning up temporary shuffle output file\n";
        exit(-1);
    }

    return resultAlignment;



}


/////////////////////////////////////////////////////////

/*
 * This returns a pointer to a new Alignment object that has been realigned by TCoffee
 * Note: You must delete this object once it is not needed any more.
 */
Alignment* Alignment::realignTcoffee(Tree * tree){
	//write fasta file:
	string fastaFilenameTemplate = "/tmp/fasta.XXXXXX";
	char fastaFilename[fastaFilenameTemplate.size()+1];
	strcpy(fastaFilename, fastaFilenameTemplate.c_str());

	int unique_fd;
	FILE * unique_file;

	if ((unique_fd = mkstemp(fastaFilename)) == -1 ){
		cerr << "Could not create temporary clustal file for shuffling\n";
		exit(-1);
	}

	if((unique_file = fdopen(unique_fd, "w")) == NULL){
		cerr << "Could not open temporary clustal file for shuffling\n";
		exit(-1);
	}

	fputs(FastaFormat(false).c_str(), unique_file);
	fclose(unique_file);

	//make a temp file for output:
	string outTemplate = "/tmp/realign_out.XXXXXX";
	char outFilename[outTemplate.size()+1];
	strcpy(outFilename, outTemplate.c_str());

	if ((unique_fd = mkstemp(outFilename)) == -1 ){
		cerr << "Could not create temporary out file for realignment\n";
		exit(-1);
	}

	if((unique_file = fdopen(unique_fd, "w")) == NULL){
		cerr << "Could not open temporary out file for realignment\n";
		exit(-1);
	}
	string temp_contents = "TEMP!\n"; // this will be replaced by TCOFFEE output
	fputs(temp_contents.c_str(), unique_file);
	fclose(unique_file);

	//make a temp file for output:
	string treeTemplate = "/tmp/tree_out.XXXXXX";
	char treeFilename[treeTemplate.size()+1];
	strcpy(treeFilename, treeTemplate.c_str());

	if ((unique_fd = mkstemp(treeFilename)) == -1 ){
		cerr << "Could not create temporary tree file for realignment\n";
		exit(-1);
	}

	if((unique_file = fdopen(unique_fd, "w")) == NULL){
		cerr << "Could not open temporary out file for realignment\n";
		exit(-1);
	}
	//write tree to file
	fputs(tree->newickString().c_str(), unique_file);
	fclose(unique_file);

	//run realignment:
	string command = TCOFFEE_LOC + " " + string(fastaFilename) + " -outfile " + string(outFilename) + " -usetree " + string(treeFilename) + " -output fasta -n_core 1 -outorder input -quiet";
	if(system(command.c_str()) != 0){
		cerr << "Error: t_coffee encountered an error\n";
		exit(-1);
	}

	ifstream realignFile(outFilename);

	if(!realignFile.is_open()){
		cerr << "Error: cannot open realignment fasta file " << string(outFilename) << endl;
		exit(-1);
	}

	string line;
	vector<string> realignment;
	vector<string> r_names;
	string seq;
	bool found_seq = false;

	while(realignFile.good()){
		getline(realignFile, line);
		Utilities::trimSpaces(line);
		//skip blank lines
		if(line.length() == 0){
			continue;
		}
		if(line.at(0) == '>'){
			if(found_seq){
				//cout << "Sequence Found: " << *seq << endl;
				realignment.push_back(seq);
			}
			line = line.substr(1);
			r_names.push_back(line);
			found_seq = true;
			seq = "";
		}
		else{
			seq = seq + line;
		}
	}

	realignFile.close();
	realignment.push_back(seq);

	assert(realignment.size() == r_names.size());
	assert(realignment.size() == alignedSeqs.size());

	for(unsigned int j = 0; j < r_names.size(); j++){
		assert(r_names[j].compare(seqNames[j]) == 0);
		assert(realignment[j].length() == realignment[0].length());
	}
	Alignment* resultAlignment = new Alignment(this->seqNames, realignment, this->codonAnnotation);
	resultAlignment->realignCodonAnnotation(this->alignedSeqs[0]);

	//clean up files
	command = "rm -f " + string(outFilename) + " " + string(fastaFilename) + " " + string(treeFilename) ;
	if(system(command.c_str()) != 0){
		cerr << "Error cleaning up temporary shuffle output file\n";
		exit(-1);
	}

	return resultAlignment;
}

double Alignment::canonicalBPPercent(int pos5, int pos3){

	if(consensusBP == NULL){
		consensusBP = new UTMatrix(this->length);

		for(unsigned int i = 0; i < this->length; i++){
			for(unsigned int j = i; j < this->length; j++){
				consensusBP->set(i, j, -1);
			}
		}
	}

	if(consensusBP->at(pos5, pos3) < 0){
		int total = 0;
		int cbp = 0;
		double cbp_percent;

		for(unsigned int i = 0; i < alignedSeqs.size(); i++){
			char a,b;
			a = alignedSeqs[i].at(pos5);
			b = alignedSeqs[i].at(pos3);

			total++;
			if(Utilities::validBP(a, b)){
				cbp++;
			}
		}
		assert(total > 0);
		cbp_percent = cbp / (double) total;
		consensusBP->set(pos5, pos3, cbp_percent);
		return cbp_percent;
	}
	else{
		return consensusBP->at(pos5, pos3);
	}

}

double Alignment::getSeqCons(unsigned int pos){
	if (seqCons == NULL){
		seqCons = new vector<double>();
		for(unsigned int i = 0; i < this->length; i++){
			seqCons->push_back(-1);
		}
	}
	if(seqCons->at(pos) < 0){

		int total = 0;
		int matching = 0;
		for(unsigned int i = 0; i < alignedSeqs.size() - 1; i++){
			for(unsigned int j = i + 1; j < alignedSeqs.size(); j++){
				if(toupper(alignedSeqs[i].at(pos)) == toupper(alignedSeqs[j].at(pos))){
					matching++;
				}
				total++;
			}
		}
		(*seqCons)[pos] = matching / (double) total;
		//(*seqCons)[pos] = seqConservationAtPosition(pos);
	}
	return seqCons->at(pos);

}


double Alignment::getCovariance(int pos5, int pos3){
	int denom = 0;
	int covariance = 0;
	for(unsigned int i = 0; i < alignedSeqs.size() - 1; i++){

		for(unsigned int j = i+1; j < alignedSeqs.size(); j++){
			int hamming = 0;
			char ia, ib, ja, jb;
			ia = toupper(alignedSeqs[i].at(pos5));
			ib = toupper(alignedSeqs[i].at(pos3));

			ja = toupper(alignedSeqs[j].at(pos5));
			jb = toupper(alignedSeqs[j].at(pos3));
			if(ia != ja){
				hamming++;
			}
			if(ib != jb){
				hamming++;
			}

			if(hamming > 0){
				if(Utilities::validBP(ia, ib) && Utilities::validBP(ja,jb)){
					covariance += hamming;
				}
				else{
					covariance -= hamming;
				}
			}
			denom++;
		}
	}
	return covariance / (double)denom;
}

string Alignment::FastaFormat(bool aligned) const{
	string out = "";
	for(unsigned int i = 0; i < alignedSeqs.size(); i++){
		out += ">" + seqNames[i] + "\n";
		if(aligned){
			out += alignedSeqs[i] + "\n";
		}
		else{
			out += seqs[i] + "\n";
		}
	}
	return out;
}

// Realigns the old codon annotation given a new sequence
// Precondition: old_seq is the first sequence in the alignment prior to realignment
void Alignment::realignCodonAnnotation(const string & old_seq) {
	// cout << old_seq << endl;
	// cout << old_cod << endl;
	// cout << new_seq << endl;

	string old_cod = this->codonAnnotation;
	string new_seq = this->alignedSeqs[0];

	assert(old_seq.length() == old_cod.length());
	int new_pos = 0;
	string new_cod = "";
	char addition = (old_cod[0] == '0') ? '0' : '3';
	bool keep = false;
	for(size_t i = 0; i < old_seq.length(); ++i) {

		if(old_seq[i] == new_seq[new_pos]) {
			new_cod += old_cod[i];
			++new_pos;
			keep = false;
		} else if (new_seq[new_pos] == '-') {
			new_cod += addition;
			++new_pos;
			--i;
		} // old_seq[i] == '-', do nothing

		if(i == old_cod.length() - 1) {
			addition = (old_cod[i] == '0') ? '0' : '3';
		} else {
			// cout << i << "(" << old_cod[i] << ", " << old_cod[i + 1] << "), ";
			if (old_cod[i] == '0' && old_cod[i + 1] == '0') {
				addition = '0';
			} else if (old_cod[i] != '0' && old_cod[i + 1] != '0') {
				addition = '3';
			} else if (!keep){
				addition = (rand() % 2) ? '0' : '3';
				keep = true;
			}
		}
	}
	new_cod += string(new_seq.length() - new_pos, addition);
	this->codonAnnotation = new_cod;
}


