/*
 * Copyright: Nicholas P Wiebe (2008-2009) and Irmtraud M Meyer (2008-2009)
 * License: licensed under the GNU General Public License version 3 (GPLv3)
 * Alignment.h
 *
 * Class for storing an alignment, and associated information
 * Stores sequences in as aligned and unaligned (gapless) strings
 * Also stores the consensus structure for the alignment, and labels
 * the consensus helices of the structure.
 *
 * The method alignmentGStats finds and stores all competing helices in the sequences
 * of the alignment. Many of the other methods in this class require it to be run first.
 */

#ifndef ALIGNMENT_H_
#define ALIGNMENT_H_

#include <set>
#include <string>
#include <vector>

#include "Helix.h"
#include "HelixFinder.h"
#include "UTMatrix.h"
#include "LikelihoodCache.h"

//////////////////////////////////////////////
#include "phy_files/utils.h"
#include "phy_files/utilsLinAlg.h"
#include "phy_files/GrammarIO.h"
#include "phy_files/GrammarWrap.h"
#include "phy_files/EmitWrapIO.h"

#include "phy_files/PhyDef.h"
#include "phy_files/DiscreteFactorGraph.h"
#include "phy_files/DfgIO.h"
#include "phy_files/TRCTMCTree.h"
#include "phy_files/PhyIO.h"
//////////////////////////////////////////////

//Some of these are unused, I think...
//#define MAX_SEQ_L  4500    // maximal sequence length
//#define MAX_WORD_L  100    // maximal length of a char* word
//#define MAX_LINE_L 1000    // maximal length of a line
#define MIN_DIST      3    // minimal sequence distance between two base-pairing letters

using namespace std;

//forward declaration... Maybe this is avoidable?
class Tree;
class LikelihoodCache;

class Alignment {
public:

	/*
	 * For this constructor, the structure must be present in dot bracket notation as the last entry
	 * of the fasta alignment file
	 */
	Alignment(string& filename);

	//alignment file is an aligned fasta file
	//struct file is a sequence file
	Alignment(string& filename, string& codonFilename);

	//create an alignment from vectors specifying sequence names, aligned sequences, and codon annotation
	Alignment(vector<string> names, vector<string> alignment, string codons);

    //create an alignment after realignment with shape data
    Alignment(vector<string> names, vector<string> alignment, string codon, string shapefile, bool cor);

    //shape file has 3 columns: reference sequence positions, bin values, nucleotides
    Alignment(string & filename, string & shapeFilename, bool & cor);

	// creates an alignment from the concatenation of two alignments. by jeffp
	Alignment(Alignment* a, Alignment* b);

    //blank constructor - do nothing   (Note: called implicitly when ShuffledAlignment objects are created)
    Alignment();

    virtual ~Alignment();

	/**
	 * various methods for printing information about the alignment, mostly for debugging
	 * purposes.
	 */
	string alignmentString();
	string columnString(int pos);
	string seqString();
	string codonString();

	/**
	 * gets the alignment in Fasta format
	 * @param aligned if true, return the gapped sequences (FastaAln format)
	 *
	 */
	string FastaFormat(bool aligned = true) const;

	/**
	 * returns fraction of canonical bp for basepair at pos5-pos3
	 */
	double canonicalBPPercent(int pos5, int pos3);

	/**
	 * returns covariance statistic for basepair at pos5-pos3
	 */
	double getCovariance(int pos5, int pos3);

	/**
	 * returns the sequence conservation at specified alignment position
	 * as measured by mean percent identity
	 */
	double getSeqCons(unsigned int pos);

	/**
	 * Calculates the logLikelihood for a single pair of (alignment) positions
	 * If this has been calculated before, then this is a simple table lookup
	 * If not, then it calculates likelihoods under the paired and singleton models
	 * of evolution using the felsenstein algorithm (O(N), where N is the number of sequences
	 * in the alignment).
	 *
	 * Precondition: You must call initializeCache on the alignment before calling any of the
	 * likelihood methods.
	 */
	double logLikelihood(int pos5, int pos3, Tree * tree);
	virtual double logPairedLikelihood(int pos5, int pos3, Tree * tree);
	virtual double logUnpairedLikelihood(int pos, Tree * tree);
    /////////////////////////////////////////////////////////
	virtual double logStackLikelihood(int pos5, int pos3, int index_pos5, int index_pos3);
    virtual double logSingleLikelihood(int pos, int index_pos);
    virtual double logPairLikelihood(int pos5, int pos3,int index_pos5,int index_pos3);
    virtual double logLikelihoodPair(int pos5, int pos3, Tree * tree);
    virtual double logLikelihoodShape(int pos5, int pos3,int index_pos5,int index_pos3);
    virtual double logLikelihoodShapePair(int pos5, int pos3,int index_pos5,int index_pos3);
    /////////////////////////////////////////////////////////

	void initializeCache();

	static bool no_cache;

	/**
	 * Returns a new alignment object that has been realigned using TCoffee
	 * Note: You must delete this object once finished using it
	 */
	Alignment* realignTcoffee(Tree * tree);
    //////////////////////////////////////////////
    Alignment* realignTcoffeeShape(Tree * tree);
    //////////////////////////////////////////////

	//for realignTcoffee:
	static const string TCOFFEE_LOC;
	static int minStemLength;
	static int minLoopLength;

	//Data structures:
	int length; // length of sequences
	vector<string> seqNames; //vector of sequence names
	vector<string> alignedSeqs; //vector of aligned sequences (with gaps)
	vector<string> seqs; //vector of ungapped sequences

    vector<int> shapePos; //has the MSA mapped reference sequence positions from 'seq2AlignmentMap' vector.
    vector<int> binValue; //has the corresponding bin values of MSA mapped reference sequence positions
    string refSeq; //stores the reference sequence

    vector<int> shapePosCut;
    vector<int> binValueCut;
    string refSeqCut;
    int refSeqID;
    vector<double> singe_likelihoods;
    boost::numeric::ublas::matrix<double> stack_mat;
    boost::numeric::ublas::matrix<double> pair_mat;
    static double multiplier_shape;
    static double factor_fels;
    static double factor_shape;
    static string path_model;
    static bool realign_cor;
    static string realign_shapename;

	vector<vector<int>* > seq2AlignmentMap; //mapping of sequence positions to alignment positions for each (ungapped) sequence
	vector<vector<int>* > alignment2SeqMap; //mapping of alignment position to sequence position for each (ungapped) sequence

	string codonAnnotation; // string of codon annotations, consisting of {0, 1, 2, 3}
	bool annotated;
	LikelihoodCache* likelihoods;




protected:

    string refSeqStore; //stores the reference sequence
    void readShape(string& filename, vector<vector<int>* >& seq2AlignmentMap);
    void fillShape(vector <int> tempShapePos, vector <int> tempBinValue, string refSeq);
    void makeCut(vector<int> shapePos, vector<int> binValue, string refSeq);
    void genAma(vector<int> shapePosCut, vector<int> binValueCut, string refSeqCut, bool cor);
    void calculateDFGmodels(string path_model,vector<int> shapePosCut,vector<int> binValueCut, string refSeqCut);

    string ama_str;
    string name_rand;
//    void genRandomName(const int len);
//    void writeFilesAmaDfg(string name_rand, bool cor);
//    void DfgLikelihoods(string name_rand);
    void corCut(vector<double> singe_likelihoods,boost::numeric::ublas::matrix<double> stack_mat,
                           vector<int> shapePos, vector<int> binValue, string refSeq);
    void upperMatrixCopy(boost::numeric::ublas::matrix<double> stack_mat);


    void readAlignment(string& filename);

    /*
     * Parses a dot bracket string to fill alignedStruct and seqStructs data structures
     */
    void parseDotBracket(string & dotBracketString);


    void initializeAlignmentMap();


	void realignCodonAnnotation(const string & old_seq);

	/**
	 * Log-odds matrices:
	 * look-up table to avoid recalculating log-odds scores for columns
	 * If any entry is > 0, that means it has yet to be calculated
	 */
	//vector<vector<double>* > felsDoubles;  //changed by Casper from double to float..nope doesn't help memory issue. Double is 8 bytes, float is 4 bytes...maybe try to use short int...nope doesn't help. the first alignment of the 2 for interactions get through, but not the second. and besdies, doesn't make sense to use short int anyway
	//vector<double> felsSingles;            //changed by Casper from double to float


	vector<double> * seqCons;//stores sequence conservation score for each column
	UTMatrix * consensusBP;

	bool hasCache;
};

#endif /* ALIGNMENT_H_ */
