/*
 * Copyright: Nicholas P Wiebe (2008-2009) and Irmtraud M Meyer (2008-2009)
 * License: licensed under the GNU General Public License version 3 (GPLv3)
 * HelixFinder.h
 *
 * Finds and stores helices in a given alignment
 */

#ifndef HELIXFINDER_H_
#define HELIXFINDER_H_

#include <vector>
#include <list>
#include <fstream>
#include "AlignedHelix.h"
#include "Tree.h"

using namespace std;

class Alignment;
class Tree;
class AlignedHelix;

enum Realigner { NO_REALIGN, TCOFFEE };

class HelixFinder {
public:
	HelixFinder(Alignment * alignment_);
	virtual ~HelixFinder();

	void findAllHelicesOld();
	void findAllHelices();
	void findAllInteractingHelices(const Alignment & alignmentA, const Alignment & alignmentB);

	void readHelix(string filename);

	/**
	 *  removes all helices that do not overlap a specified boundary index
	 *  where the boundary represents the position between concatenated alignments
	 *  by jeffp
	 */
	void filterForABHelices(int boundaryIndex);

	/**
	 * calculate pvalues for log ratios of all found helices,
	 * by generating randomized alignments to simulate null distribution.
	 *
	 * @param randomSamples Number of randomized alignments to generate
	 * @param doPvalues if false, don't do pvalue calculation - just fill that column with zeros
	 * @param PValueRange Desired range of most significant p-values to be calculated
	 * @param fast if true, calculate p-values with accelerated heuristics
	 */
	void allHelicesPvalueTable(Tree * tree);

	static void allHelicesInteractionsPvalueTable(string & outputFilePrefix, string & helixLogPrefix,
			Tree * tree, Alignment* alignA, Alignment* alignB, HelixFinder &hfOverlap);

	/*
	static unsigned int processRandomTrial(HelixFinder & randomHelices, HelixFinder & realHelices,
			vector<double> & randLogLikes, vector<double> & allRandLogLikes,
			vector<unsigned int> & lowerHelixCount, Tree * tree, bool writeHelixLog, int numTrials, int nullHelixCount, double & maxLog,
			std::ofstream* log); */

	//added by Casper
	static unsigned int processRandomTrial(HelixFinder & randomHelices, HelixFinder & realHelices,
			vector<double> & randLogLikes, vector<double> & allRandLogLikes,
			vector<unsigned int> & lowerHelixCount, Tree * tree, bool writeHelixLog, int numTrials, int nullHelixCount, double & maxLog,
			std::ofstream* log, int boundaryIndex, int a_b_ab);

	//depreciated
	static void writeHelixData(string & filename, vector<double> & newPvalues, vector<double> & allRandLogLikes,
			vector<AlignedHelix> & helices, Alignment & alignment, Tree * tree, double & maxLog, int alignALength = -1);

	//added by Casper
	static void writeHelixData(string & filename, vector<double> & newPvalues, vector<double> & allRandLogLikes,
			vector<AlignedHelix> & helices, Alignment* alignment, Tree * tree, double & maxLog, Alignment* alignment2, int alignALength = -1, int boundaryIndex=-1, int a_b_ab=-1);

	void writeHelixDataMinimal(string filename, Tree * tree);

	// debug
	void printHelices();
	
	/** EXPERIMENTAL
	 * This is a relatively odd concept. Rather than finding all helices at the level of
	 * individual sequences, it first calculates the log-likelihood ratio of all pairs of columns
	 * and then looks for columns that have a log likelihood above the seed threshold. These columns
	 * are used as the seeds of helices, and are expanded by adding base-pairs inner or outer to
	 * the seed pair until the log-likelihood of the helix falls below a certain threshold.
	 * Then, all helices below the minimum helix length are discarded.
	 *
	 * I can't remember if this function has been thoroughly tested, so it may not work as advertised.
	 * Nevertheless, it's an interesting concept, maybe? The algorithm based on a
	 * a very basic method of image segmentation. -njwiebe
	 */
	void findAllHelicesGrow(Tree * tree, double seedThreshold, double growThreshold);

	vector<AlignedHelix> helices;

	static Realigner realign;
	static bool print_struct;
	static bool debug_verbose;
	static bool doPvalues;
	static double integration_cutoff; // maximum value under which integration is conducted in p-value calculation
	static bool sample_less;
	static bool integrate;
	static float pValueRange;
    //////////////////////////////////
    static bool first_seq_only;
    static string log_distribution_file;
    //////////////////////////////////
	//added by Casper
	static float logratiothreshold;
	static float pvaluethreshold;

private:
	Alignment * alignment; //reference instead of pointer?

	/**
	 * find all helices in a specific sequence from the alignment
	 * (not added to helices vector, just to helicesByOuterBP vector)
	 */
	void findAllHelices(int seqIndex, vector<list<AlignedHelix> > & helicesByOuterBP);

	/**
	 * converts a pair of positions to an index in a flattened UT matrix, as used in helicesByOuterBP
	 */
	int convert(int pos5, int pos3);

};

#endif /* HELIXFINDER_H_ */
