/*
 * Copyright: Nicholas P Wiebe (2008-2009) and Irmtraud M Meyer (2008-2009)
 * License: licensed under the GNU General Public License version 3 (GPLv3)
 */

#include "ShuffledAlignment.h"

#include <cassert>
#include <list>
#include <vector>
#include <fstream>
#include <iostream>
#include <cstdlib>
#include <cstring>
#include <algorithm>

#include "Utilities.h"

ShuffledAlignment::ShuffledAlignment(Alignment* a){

	seqNames = a->seqNames;
	alignedSeqs = a->alignedSeqs;
	codonAnnotation = a->codonAnnotation;
	annotated = a->annotated;
	length = a->length;
	likelihoods = a->likelihoods;
	columnMap = new int[a->length];
    //////////////////////////////////////////////////////////////
    bool cor = false;
//    this->genRandomName(100);
    refSeqID = a->refSeqID;
    //generate aligned length bins for future
    for(unsigned int i = 0; i < alignedSeqs[0].length(); ++i) {
        this->binValueAln.push_back(-1);
    }

    vector<int> shapePosTmp = a->shapePos;
    vector<int> binValueTmp = a->binValue;

    for(int i = 0; i < shapePosTmp.size(); ++i) {
        this->binValueAln[shapePosTmp.at(i)] = binValueTmp[i];
    }
    shuffledBinAln.clear();
    //////////////////////////////////////////////////////////////

	if (annotated) {
		codonShuffle();
	} else {
		shuffleWindow(alignedSeqs, 0);
	}

	this->initializeAlignmentMap();

    /////////////////////////////////////////////////////////////////////////////
    //now we need to return back shuffledBinAln to shuffledBin
    this->shapePos.clear();
    this->binValue.clear();
    this->shapePos.reserve(this->seq2AlignmentMap[refSeqID]->size());
    this->binValue.reserve(this->seq2AlignmentMap[refSeqID]->size());
    this->refSeq = this->seqs[refSeqID];

    for(unsigned int i = 0; i < this->seq2AlignmentMap[refSeqID]->size(); ++i) {
        this->shapePos.push_back(this->seq2AlignmentMap[refSeqID]->at(i));
        this->binValue.push_back(this->shuffledBinAln[this->seq2AlignmentMap[refSeqID]->at(i)]);
    }

//    cout << "this->binValue:" << endl;
//    for (auto i = this->binValue.begin(); i != this->binValue.end(); ++i)
//        std::cout << *i << '\t';
//    cout << "\n:" << endl;
//
//    cout << "a->binValue:" << endl;
//    for (auto i = binValueTmp.begin(); i != binValueTmp.end(); ++i)
//        std::cout << *i << '\t';
//    cout << "\n:" << endl;

//    cout << "this->shapePos:" << endl;
//    for (auto i = this->shapePos.begin(); i != this->shapePos.end(); ++i)
//        std::cout << *i << '\t';
//    cout << "\n:" << endl;
//
//    cout << "a->shapePos:" << endl;
//    for (auto i = shapePosTmp.begin(); i != shapePosTmp.end(); ++i)
//        std::cout << *i << '\t';
//    cout << "\n:" << endl;

    /////////////////////////////////////////////////////////////////////////////

	// sanity check that the column map is constructed correctly. Can be removed once verified.
	for (int i = 0; i < length; i++) {
		assert (this->columnString(i).compare(a->columnString(this->columnMap[i])) == 0);
	}
    /////////////////////////////////////////////////////////////////////////////
//    this->makeCut(this->shapePos,this->binValue,this->refSeq);
//    this->genAma(this->shapePos,this->binValue,this->refSeq,cor);
//    this->writeFilesAmaDfg(this->name_rand,cor);
//    this->DfgLikelihoods(this->name_rand);
//    this->upperMatrixCopy(this->stack_mat);

//    this->makeCut(this->shapePos,this->binValue,this->refSeq);
//    this->calculateDFGmodels(this->path_model, this->shapePosCut, this->binValueCut, this->refSeqCut);
//    this->corCut(this->singe_likelihoods,stack_mat,shapePos,binValue,refSeq);
//    this->upperMatrixCopy(this->stack_mat);

    this->binValueCut = this->binValue;
    this->shapePosCut = this->shapePos;
    this->refSeqCut = this->refSeq;

//    makeCut(shapePos,binValue,refSeq);
    calculateDFGmodels(this->path_model, this->shapePosCut, this->binValueCut, this->refSeqCut);
//    corCut(singe_likelihoods,stack_mat,shapePos,binValue,refSeq);
    upperMatrixCopy(this->stack_mat);

    /////////////////////////////////////////////////////////////////////////////

}

/*
 *  Concatenates two shuffled alignments while preserving its caching ability
 *  Calls constructor super class Alignment(*Alignment, *Alignment)
 */
ShuffledAlignment::ShuffledAlignment(ShuffledAlignment* subAlignmentA, ShuffledAlignment* subAlignmentB, Alignment* concatenated)
			: Alignment(subAlignmentA, subAlignmentB) {
	this->likelihoods = concatenated->likelihoods;

	this->columnMap = new int[subAlignmentA->length + subAlignmentB->length];
	for (int i = 0; i < subAlignmentA->length; i++) {
		this->columnMap[i] = subAlignmentA->columnMap[i];
	}
	for (int i = 0; i < subAlignmentB->length; i++) {
		this->columnMap[i + subAlignmentA->length] = subAlignmentB->columnMap[i] + subAlignmentA->length;
	}

	// sanity check that the columnMap is constructed properly. Could be removed once verified
	for (int i = 0; i < length; i++) {
		assert(this->columnString(i).compare(concatenated->columnString(this->columnMap[i])) == 0);
	}
}

ShuffledAlignment::~ShuffledAlignment() {
	delete columnMap;
}


double ShuffledAlignment::logPairedLikelihood(int pos5, int pos3, Tree * tree){
	int mappedPos5 = this->columnMap[pos5];
	int mappedPos3 = this->columnMap[pos3];
	if (hasCache && likelihoods->isCached(mappedPos5, mappedPos3)) {
		return likelihoods->getCached(mappedPos5, mappedPos3);
	}
	else {
		double likelihood = tree->calcLogFelsPaired(*this, pos5, pos3);
		if (hasCache) likelihoods->setCached(mappedPos5, mappedPos3, likelihood);
		return likelihood;
	}
}

double ShuffledAlignment::logUnpairedLikelihood(int pos, Tree * tree){
	int mappedPos = this->columnMap[pos];
	if (hasCache && likelihoods->isCached(mappedPos)) {
		return likelihoods->getCached(mappedPos);
	}
	else {
		double likelihood = tree->calcLogFelsUnpaired(*this, pos);
		if (hasCache) likelihoods->setCached(mappedPos, likelihood);
		return likelihood;
	}
}



vector<string>  ShuffledAlignment::sliceAlnByColumn(vector<string>& alignedSeqs, const int &sliceStart, const int &sliceEnd) {
	vector<string> slice;
	for(unsigned i = 0; i < alignedSeqs.size(); ++i) {
		slice.push_back(alignedSeqs[i].substr(sliceStart, sliceEnd - sliceStart));
	}
	return slice;
}

vector<string> ShuffledAlignment::shuffleAln(vector<string>& aln, int start) {
	unsigned int maxRow = aln.size();
	unsigned int maxCol = aln[0].length();

	vector<string> masks;
	map<string, vector<int> > bins;

	for(unsigned int currCol = 0; currCol < maxCol; ++currCol) {
		string mask;

		// Creates gap mask for column
		for(unsigned int currRow = 0; currRow < maxRow; ++currRow) {
			if(aln[currRow].at(currCol) == '-') {
				mask.append("-");
			} else {
				mask.append("X");
			}
		}

		// Calculate pairwise identity for each column
		int match = 0;
		int pairs = 0;

		for(unsigned int i = 0; i < maxRow; ++i) {
			for(unsigned int j = i + 1; j < maxRow; ++j) {
				char charI = toupper(aln[i][currCol]);
				char charJ = toupper(aln[j][currCol]);

				// cout << charI << " x " << charJ << endl;

				if (charI != '-' and charJ != '-') {
					if(charI == charJ) {
						++match;
					}
					++pairs;
				}
			}
		}

		char id[4];
		if (pairs > 0) {
			sprintf(id, "%.1f", (double)match / pairs);
		} else {
			sprintf(id, "%.0f", 1.00);
		}

		mask.append(id, 3);
		masks.push_back(mask);
		bins[mask].push_back(currCol);
	}

	// Shuffle bins with Fisher-Yates algorithm

	for (map<string, vector<int> >::iterator it = bins.begin() ; it != bins.end(); it++ ) {
		random_shuffle((*it).second.begin(), (*it).second.end());

		/*
		// Confirmed identical implementation as RNAz
		// Suspect to modulo bias?
		vector<int> *array = &(*it).second;
		for (int i = array->size() - 1; i > 0; --i) {
			int j = rand() % (i + 1);
			// int j = 1000 % (i + 1);
			int temp = (*array)[i];
			(*array)[i] = (*array)[j];
			(*array)[j] = temp;

		}
		*/
	}

	// Assemble shuffled alignment
	vector<string> shuffledAln;
	for(unsigned int i = 0; i < maxRow; ++i) {
		shuffledAln.push_back(string(maxCol, '*'));
	}
    ////////////////////////////////////////
    this->shuffledBinAln.reserve(this->binValueAln.size());
    for(unsigned int i = 0; i < this->binValueAln.size(); ++i) {
        this->shuffledBinAln.push_back(-1);
    }
    ////////////////////////////////////////

	for(unsigned int currCol = 0; currCol < maxCol; ++currCol) {
		int randomCol = bins[masks[currCol]].back();
		bins[masks[currCol]].pop_back();
		this->columnMap[start + currCol] = start + randomCol;

		for(unsigned int currRow = 0; currRow < maxRow; ++currRow) {
			shuffledAln[currRow][currCol] = aln[currRow][randomCol];
		}
        /////////////////////////////////////////////////////////////
        this->shuffledBinAln[currCol] = this->binValueAln[randomCol];
		/////////////////////////////////////////////////////////////
	}
	return shuffledAln;
}

vector<string> ShuffledAlignment::unbinnedShuffle(vector<string>& aln, int start) {
	unsigned int maxRow = aln.size();
	unsigned int maxCol = aln[0].length();

	vector<int> bins;

	for(unsigned int currCol = 0; currCol < maxCol; ++currCol) {
		bins.push_back(currCol);
	}

	// Shuffle bins with Fisher-Yates algorithm

	random_shuffle(bins.begin(), bins.end());

	// Assemble shuffled alignment
	vector<string> shuffledAln;
	for(unsigned int i = 0; i < maxRow; ++i) {
		shuffledAln.push_back(string(maxCol, '*'));
	}

	for(unsigned int currCol = 0; currCol < maxCol; ++currCol) {
		int randomCol = bins.back();
		bins.pop_back();
		this->columnMap[start + currCol] = start + randomCol;

		for(unsigned int currRow = 0; currRow < maxRow; ++currRow) {
			shuffledAln[currRow][currCol] = aln[currRow][randomCol];
		}
	}
	return shuffledAln;
}


vector<string> ShuffledAlignment::shuffleAlnStatic(vector<string>& aln, int start) {
	unsigned int maxRow = aln.size();
	unsigned int maxCol = aln[0].length();

	vector<string> masks;
	map<string, vector<int> > bins;

	for(unsigned int currCol = 0; currCol < maxCol; ++currCol) {
		string mask;

		// Creates gap mask for column
		for(unsigned int currRow = 0; currRow < maxRow; ++currRow) {
			if(aln[currRow].at(currCol) == '-') {
				mask.append("-");
			} else {
				mask.append("X");
			}
		}

		// Calculate pairwise identity for each column
		int match = 0;
		int pairs = 0;

		for(unsigned int i = 0; i < maxRow; ++i) {
			for(unsigned int j = i + 1; j < maxRow; ++j) {
				char charI = toupper(aln[i][currCol]);
				char charJ = toupper(aln[j][currCol]);

				// cout << charI << " x " << charJ << endl;

				if (charI != '-' and charJ != '-') {
					if(charI == charJ) {
						++match;
					}
					++pairs;
				}
			}
		}

		char id[4];
		if (pairs > 0) {
			sprintf(id, "%.1f", (double)match / pairs);
		} else {
			sprintf(id, "%.0f", 1.00);
		}

		mask.append(id, 3);
		masks.push_back(mask);
		bins[mask].push_back(currCol);
	}

	// Shuffle bins with Fisher-Yates algorithm

	for (map<string, vector<int> >::iterator it = bins.begin() ; it != bins.end(); it++ ) {
		random_shuffle((*it).second.begin(), (*it).second.end());
	}

	// Assemble shuffled alignment
	vector<string> shuffledAln;
	for(unsigned int i = 0; i < maxRow; ++i) {
		shuffledAln.push_back(string(maxCol, '*'));
	}

	for(unsigned int currCol = 0; currCol < maxCol; ++currCol) {
		int randomCol = bins[masks[currCol]].back();
		bins[masks[currCol]].pop_back();
		for(unsigned int currRow = 0; currRow < maxRow; ++currRow) {
			shuffledAln[currRow][currCol] = aln[currRow][randomCol];
		}
	}
	return shuffledAln;
}

vector<string> ShuffledAlignment::unbinnedShuffleStatic(vector<string>& aln, int start) {
	unsigned int maxRow = aln.size();
	unsigned int maxCol = aln[0].length();

	vector<int> bins;

	for(unsigned int currCol = 0; currCol < maxCol; ++currCol) {
		bins.push_back(currCol);
	}

	// Shuffle bins with Fisher-Yates algorithm
	random_shuffle(bins.begin(), bins.end());

	// Assemble shuffled alignment
	vector<string> shuffledAln;
	for(unsigned int i = 0; i < maxRow; ++i) {
		shuffledAln.push_back(string(maxCol, '*'));
	}

	for(unsigned int currCol = 0; currCol < maxCol; ++currCol) {
		int randomCol = bins.back();
		bins.pop_back();
		for(unsigned int currRow = 0; currRow < maxRow; ++currRow) {
			shuffledAln[currRow][currCol] = aln[currRow][randomCol];
		}
	}
	return shuffledAln;
}

// All following arrays are simply to support the randomizeCodon function
const char int_to_char[] = {'A', 'U', 'G', 'C'};

const int char_to_int[] = {
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, // 0   - 9
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, // 10  - 19
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, // 20  - 29
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, // 30  - 39
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, // 40  - 49
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, // 50  - 59
	4, 4, 4, 4, 4, 0, 4, 3, 4, 4, // 60  - 69, 65 = A (0), 67 = C (3)
	4, 2, 4, 4, 4, 4, 4, 4, 4, 4, // 70  - 79, 74 = G (2)
	4, 4, 4, 4, 4, 1, 4, 4, 4, 4, // 80  - 89, 85 = U (1)
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, // 90  - 99
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, // 100 - 109
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, // 110 - 119
	4, 4, 4, 4, 4, 4, 4, 4		  // 120 - 127
};

// codon: AAA = 0, AAU = 1, AAC = 2, AAG = 3, etc., CCC = 63
// amino: 0 = A (Ala), 1 = B (Unused), 2 = C (Cys), etc., 23 = X = (STOP)
const int codon_to_amino[] = {
	10, 13, 10, 13,  8,  8, 12,  8, 17, 18,
	17, 18, 19, 19, 19, 19, 23, 24, 23, 24,
	11,  5, 11,  5, 23,  2, 22,  2, 18, 18,
	18, 18,  4,  3,  4,  3, 21, 21, 21, 21,
	 6,  6,  6,  6,  0,  0 , 0,  0, 16,  7,
	16,  7, 11, 11, 11, 11, 17, 17, 17, 17,
	15, 15, 15, 15
};

// The first value of each row is the number of entries following it
// The following entries correspond to encoded codons
const int amino_to_codons[26][7] = {
	{4, 44, 45, 46, 47},			// Ala (A)
	{0},							// Unused (B)
	{2, 25, 27},					// Cys (C)
	{2, 33, 35},					// Asp (D)
	{2, 32, 34},					// Glu (E)
	{2, 21, 23},					// Phe (F)
	{4, 40, 41, 42, 43},			// Gly (G)
	{2, 49, 51},					// His (H)
	{3, 4, 5, 7},					// Ile (I)
	{0},							// Unused (J)
	{2, 0, 2},						// Lys (K)
	{6, 20, 22, 52, 53, 54, 55},	// Leu (L)
	{1, 6},							// Met (M)
	{2, 1, 3},						// Asn (N)
	{0},							// Unused (O)
	{4, 60, 61, 62, 63},			// Pro (P)
	{2, 48, 50},					// Gln (Q)
	{6, 10, 56, 57, 58, 59, 8},		// Arg (R)
	{6, 11, 28, 29, 30, 31, 9},		// Ser (S)
	{4, 12, 13, 14, 15},			// Thr (T)
	{0},							// Unused (U)
	{4, 36, 37, 38, 39},			// Val (V)
	{1, 26},						// Trp (W)
	{3, 16, 18, 24},				// STOP (X)
	{2, 17, 19},					// Tyr (Y)
	{0}								// Unused (Z)
};

// Decodes codon numbers from amino_to_codons
static string num_to_codon(int number) {
	string codon = "";
	codon += int_to_char[number / 16];
	number %= 16;
	codon += int_to_char[number / 4];
	number %= 4;
	codon += int_to_char[number];
	return codon;
}

static int codon_to_num(string & codon) {
	return 	char_to_int[(int)codon[0]] *
			16 + char_to_int[(int)codon[1]] * 4 + char_to_int[(int)codon[2]];
}

string ShuffledAlignment::randomizeCodon(string codon) {
	assert(codon.size() == 3);

	int gaps = count(codon.begin(), codon.end(), '-');
	if(gaps) {
		return codon;
	} // TODO: Do something else? e.g. shuffle randomly when gap == 1?

	int number = codon_to_num(codon);
	int amino = codon_to_amino[number];
	int size = amino_to_codons[amino][0];
	//Modified by Casper: some codons, like NCC, can result in size of 0, leading to division by 0
	if(size > 0){
		int random = rand() % size + 1;
		int pick = amino_to_codons[amino][random];
		return num_to_codon(pick);
	}
	//Added by Casper: if size is 0, meaning the codon does not point to any legit amino acid, then just return the same as it is
	else if(size == 0){
		return num_to_codon(size);
	}
	//Sanity check
	else{
		cerr << "Error in ShuffledAlignment::randomizeCodon. Processing a negative value size: " << size;
		return "-1";
	}
	return "-1";
	// Equivalent call, no speed difference with 10000000 calls
	// Unknown cache/memory difference...
	// return num_to_codon(amino_to_codons[codon_to_amino[codon_to_num(codon)]][rand() % amino_to_codons[codon_to_amino[codon_to_num(codon)]][0] + 1]);
}

// Shuffles a given sequence alignment in windows of 120
void ShuffledAlignment::shuffleWindow(vector<string> & sequences, int regionStart) {
	const int window = sequences[0].length();
	const int slide = sequences[0].length();
//    const int window = 1;
//    const int slide = 1;

	int length = sequences[0].length();
	int sliceStart = 0;
	int sliceEnd = 0;

	vector<vector<string> > shuffledWindows;

	while (sliceStart < length) {
		sliceEnd = sliceStart + window;
		if (sliceEnd > length) sliceEnd = length;

		vector<string> slice = sliceAlnByColumn(sequences, sliceStart, sliceEnd);

		slice = this->shuffleAln(slice, regionStart + sliceStart);
//		slice = this->unbinnedShuffle(slice, regionStart + sliceStart);
		shuffledWindows.push_back(slice);
		sliceStart += slide;
		if(sliceEnd == length) break;
	}

	concatenateWindows(sequences, shuffledWindows);
	return;
}

void ShuffledAlignment::concatenateWindows(vector<string> & sequences,
		vector<vector<string> > shuffledWindows) {

	for(unsigned int i = 0; i < sequences.size(); ++i) {
		string assembledSeq = "";
		for(unsigned int j = 0; j < shuffledWindows.size(); ++j) {
			vector<string> slice = shuffledWindows[j];
			string piece = slice[i];
			assembledSeq += piece;
		}
		sequences[i] = assembledSeq;
	}

}

void ShuffledAlignment::codonShuffle() {
	int non_coding_start = 0;
	int non_coding_end = 0;
	bool non_coding = 0;
	int coding_start = 0;
	int coding_end = 0;

	// Split coding and non-coding windows, shuffle non-coding, store coding
	vector<vector<string> > shuffledWindows;
	vector<vector<string> > codingWindows;
	vector<int> codingWindowLocation;
	vector<int> codingWindowLength;
	string codingAnnotation;
	for(size_t i = 0; i < codonAnnotation.size(); ++i) {
		if (codonAnnotation[i] == '0' && !non_coding) { // Starting non-coding
			non_coding_start = i;
			coding_end = i;
			non_coding = 1;
			if (coding_end > 0) {
				// cout << "Dumping Coding: " << coding_start << ", " << coding_end << endl;
				vector<string> slice = sliceAlnByColumn(alignedSeqs, coding_start, coding_end);
				shuffledWindows.push_back(slice);
				codingWindows.push_back(slice);
				codingWindowLocation.push_back(shuffledWindows.size() - 1);
				codingWindowLength.push_back(slice[0].length());
				codingAnnotation += codonAnnotation.substr(coding_start, coding_end - coding_start);
			}
		} else if (codonAnnotation[i] != '0' && non_coding) { // Starting coding
			coding_start = i;
			non_coding_end = i;
			non_coding = 0;
			if (non_coding_end > 0) {
				// cout << "Dumping Non-coding: " << non_coding_start << ", " << non_coding_end << endl;
				vector<string> slice = sliceAlnByColumn(alignedSeqs, non_coding_start, non_coding_end);
				shuffleWindow(slice, non_coding_start);
				shuffledWindows.push_back(slice);
			}
		} // 0 and non_coding, 1 and coding, do nothing
	}

	// Final loop after while
	if(non_coding) {
		non_coding_end = codonAnnotation.size();
		// cout << "Dumping Non-coding: " << non_coding_start << ", " << non_coding_end << endl;
		vector<string> slice = sliceAlnByColumn(alignedSeqs, non_coding_start, non_coding_end);
		shuffleWindow(slice, non_coding_start);
		shuffledWindows.push_back(slice);
	} else {
		coding_end = codonAnnotation.size();
		// cout << "Dumping Coding: " << coding_start << ", " << coding_end << endl;
		vector<string> slice = sliceAlnByColumn(alignedSeqs, coding_start, coding_end);
		shuffledWindows.push_back(slice);
		codingWindows.push_back(slice);
		codingWindowLocation.push_back(shuffledWindows.size() - 1);
		codingWindowLength.push_back(slice[0].length());
		codingAnnotation += codonAnnotation.substr(coding_start, coding_end - coding_start);
	}

	// Concatenate all coding windows and prepare to shuffle
	vector<string> concatCodingWindows = codingWindows[0];
	concatenateWindows(concatCodingWindows, codingWindows);
	vector<int> frames = parseAnnotation(codingAnnotation);

	/*
	for(int i = 0; i < codingWindowLocation.size(); ++i) {
		cout << codingWindowLocation[i] << "(" << codingWindowLength[i] << ")" << endl;
	}

	for (int i = 0; i < concatCodingWindows.size(); ++i) {
		cout << i << ": " << concatCodingWindows[i] << endl;
	} cout << endl;

	cout << "Coding Annotation" << endl;
	cout << codingAnnotation << endl;

	for (int i = 0; i < frames.size(); ++i) {
		cout << frames[i] << ", ";
	} cout << endl;
	 */

	// Randomize codons in coding windows
	for(vector<string>::iterator seq = concatCodingWindows.begin(); seq < concatCodingWindows.end(); ++seq) {
		int pos = 0;
		string randomizedSeq;
		for(vector<int>::iterator frame = frames.begin(); frame < frames.end(); ++frame) {
			if(*frame == 0) {
				randomizedSeq += randomizeCodon((*seq).substr(pos, 3));
				// randomizedSeq += "123";
				pos += 3;
			} else {
				randomizedSeq += (*seq).substr(pos, *frame);
				pos += *frame;
			}
		}
		*seq = randomizedSeq;
	}

	// Unsplice randomized codons
	int start = 0;
	for (size_t i = 0; i < codingWindowLocation.size(); ++i) {
		int length = codingWindowLength[i];
		int window = codingWindowLocation[i];
		for(size_t j = 0; j < concatCodingWindows.size(); ++j) {
			shuffledWindows[window][j] = concatCodingWindows[j].substr(start, length);
		}
		start += length;
	}
/*
	for (int i = 0; i < concatCodingWindows.size(); ++i) {
		cout << i << ": " << concatCodingWindows[i] << endl;
	} cout << endl;
*/
	concatenateWindows(alignedSeqs, shuffledWindows);
}

// Returns a parse representation of the annotation string
// n == 0: valid triple (in order to differentiate from n == 3)
// n > 0: invalid run of n bases
vector<int> ShuffledAlignment::parseAnnotation(string codons) {
	vector<int> output;
	string input = codons;
	size_t loc = input.find("123", 0);
	size_t expected = 0;
	while(loc != string::npos) {
		if (loc == expected) {
			output.push_back(0);
		} else {
			output.push_back(loc - expected);
			output.push_back(0);
		}
		expected = loc + 3;
		loc = input.find("123", loc + 3);
	}
	if (expected != codons.length()) {
		output.push_back(codons.length() - expected);
	}

	return output;
}

